import os.path
import random
import string
import time
from collections import namedtuple
from subprocess import check_call
from subprocess import PIPE
from subprocess import Popen

import pymysql
import pytest


# To test functions that query UCB LDAP for people, we simply pick
# someone who has a few years before graduating and use their name and
# CalNet UID. Alumni eventually get kicked out of the university's
# "People" OU, so these constants have to be updated every few years
# after the aforementioned party has graduated.
TEST_PERSON_CALNET_UID = 1101587
TEST_PERSON_NAME = 'Jason Perrin'


MYSQL_TIMEOUT = 10


@pytest.yield_fixture(scope='session')
def mysqld_socket(tmpdir_factory):
    """Yield a socket to a running MySQL instance."""
    tmpdir = tmpdir_factory.mktemp('var')
    socket = tmpdir.join('socket')
    data_dir = tmpdir.join('data')
    data_dir.ensure_dir()

    check_call((
        os.path.join('/usr', 'bin', 'mysql_install_db'),
        '--no-defaults',
        '--basedir=/usr',
        '--datadir=' + data_dir.strpath,
    ))
    proc = Popen((
        os.path.join('/usr', 'sbin', 'mysqld'),
        '--no-defaults',
        '--skip-networking',
        '--lc-messages-dir', os.path.join('/usr', 'share', 'mysql'),
        '--datadir', data_dir.strpath,
        '--socket', socket.strpath,
    ))

    elapsed = 0
    step = 0.1
    while elapsed < MYSQL_TIMEOUT and not _mysql_ready(socket):
        elapsed += step
        time.sleep(step)

    try:
        yield socket
    finally:
        proc.terminate()
        proc.wait()


def _mysql_ready(socket):
    try:
        get_connection(socket)
    except pymysql.err.OperationalError:
        return False
    else:
        return True


class TemporaryMySQLDatabase(namedtuple('TemporaryMySQLDatabase', (
    'db_name',
    'mysqld_socket',
))):

    def __enter__(self):
        with self.connection(use_db=False) as c:
            c.execute('CREATE DATABASE {};'.format(self.db_name))
        return self

    def __exit__(self, type_, value, tb):
        with self.connection(use_db=False) as c:
            c.execute('DROP DATABASE {};'.format(self.db_name))

    def connection(self, use_db=True):
        kwargs = {}
        if use_db:
            kwargs['db'] = self.db_name
        return get_connection(self.mysqld_socket, **kwargs)

    def run_cli_query(self, query):
        """Run a query using the mysql CLI.

        This is useful because things like delimiters (semicolon) and complex
        statements are hard to write using PyMySQL while still being compatible
        with imports of the schema into the actual database.
        """
        mysql = Popen(
            (
                os.path.join('/usr', 'bin', 'mysql'),
                '-h', 'localhost',
                '-u', 'root',
                '--password=',
                '--database', self.db_name,
                '--socket', self.mysqld_socket.strpath,
            ),
            stdin=PIPE,
        )
        mysql.communicate(query)
        assert mysql.wait() == 0


@pytest.yield_fixture
def mysql_database(mysqld_socket):
    with TemporaryMySQLDatabase(
            db_name='test_' + ''.join(random.choice(string.ascii_lowercase) for _ in range(20)),
            mysqld_socket=mysqld_socket
    ) as d:
        yield d


def get_connection(unix_socket, **kwargs):
    return pymysql.connect(
        unix_socket=unix_socket.strpath,
        user='root',
        password='',
        cursorclass=pymysql.cursors.DictCursor,
        autocommit=True,
        charset='utf8mb4',
        **kwargs
    )



from datetime import datetime
from datetime import timedelta

import pytest

from ocflib.lab.stats import list_desktops
from ocflib.lab.stats import UtilizationProfile


def test_list_desktops():
    desktops = list_desktops()
    assert 10 < len(desktops) < 50

    assert 'eruption' in desktops
    assert 'destruction' in desktops

    assert 'death' not in desktops


@pytest.mark.xfail(reason='staff_only moved to hiera broke test')
def test_list_desktops_staff_only():
    desktops = list_desktops(public_only=True)
    assert 10 < len(desktops) < 50

    assert 'destruction' in desktops

    assert 'eruption' not in desktops
    assert 'death' not in desktops


def test_fast_slow_profiles_same():
    start = datetime(2015, 11, 23)
    end = start + timedelta(days=1)

    slow_profiles = {
        host + '.ocf.berkeley.edu': UtilizationProfile.from_hostname(host, start, end)
        for host in list_desktops()
    }
    fast_profiles = UtilizationProfile.from_hostnames(list_desktops(), start, end)

    assert set(slow_profiles.keys()) == set(fast_profiles.keys())

    for host in slow_profiles.keys():
        slow = slow_profiles[host]
        fast = fast_profiles[host]

        assert slow.hostname == fast.hostname
        assert slow.start == fast.start
        assert slow.end == fast.end
        assert slow.sessions == fast.sessions



import mock
import pytest

from ocflib.ucb.directory import calnet_uids_by_name
from ocflib.ucb.directory import name_by_calnet_uid
from tests.conftest import TEST_PERSON_CALNET_UID
from tests.conftest import TEST_PERSON_NAME


class TestNameByCalNetUID:

    @pytest.mark.parametrize('attrs,expected', [
        ({'givenName': ['Jason'], 'sn': ['Perrin']}, 'Jason Perrin'),
        ({'displayName': 'Matthew McAllister'}, 'Matthew McAllister'),
        ({}, None),
    ])
    def test_name_by_calnet_uid(self, attrs, expected):
        """This tests that the 'displayName' fallback works without hitting
        UCB LDAP, since we can't actually control entries there."""
        with mock.patch(
            'ocflib.account.search.user_attrs_ucb',
            return_value=attrs
        ):
            assert name_by_calnet_uid(0) == expected

    @pytest.mark.parametrize('uid,expected', [
        (TEST_PERSON_CALNET_UID, TEST_PERSON_NAME),
        (9999999, None),
    ])
    def test_name_by_calnet_uid_real_query(self, uid, expected):
        assert (name_by_calnet_uid(uid) or '').lower() == (expected or '').lower()


class TestCalNetUIDsByName:

    @pytest.mark.parametrize('name,expected', [
        (TEST_PERSON_NAME, [TEST_PERSON_CALNET_UID]),
        ('~~~', []),
    ])
    def test_calnet_uids_by_name(self, name, expected):
        assert calnet_uids_by_name(name) == expected



import subprocess

import mock
import pytest

from ocflib.infra.kerberos import create_kerberos_principal_with_keytab
from ocflib.infra.kerberos import get_kerberos_principal_with_keytab


class TestCreateKerberosPrincipal:

    @mock.patch('pexpect.spawn')
    def test_normal_password(self, mock_spawn):
        mock_spawn.return_value.exitstatus = 0

        create_kerberos_principal_with_keytab(
            'ggroup',
            '/some/keytab',
            'create/admin',
            password='hunter2',
        )

        mock_spawn.assert_called_once_with(
            ('/usr/bin/kadmin -K /some/keytab -p create/admin add ' +
             '--use-defaults ggroup'),
            timeout=10,
        )
        mock_spawn.return_value.sendline.assert_has_calls(
            [mock.call('hunter2'), mock.call('hunter2')],
        )

    @mock.patch('pexpect.spawn')
    def test_random_password(self, mock_spawn):
        mock_spawn.return_value.exitstatus = 0

        create_kerberos_principal_with_keytab(
            'ggroup',
            '/some/keytab',
            'create/admin',
        )

        mock_spawn.assert_called_once_with(
            ('/usr/bin/kadmin -K /some/keytab -p create/admin add ' +
             '--use-defaults ggroup'),
            timeout=10,
        )
        assert len(mock_spawn.return_value.sendline.call_args[0][0]) == 100

    @mock.patch('pexpect.spawn')
    def test_errors(self, mock_spawn):
        mock_spawn.return_value.before.decode.return_value = (
            'kadmin: kadm5_create_principal: Principal already exists'
        )
        mock_spawn.return_value.exitstatus = 1

        with pytest.raises(ValueError):
            create_kerberos_principal_with_keytab(
                'ggroup',
                '/some/keytab',
                'create/admin',
            )


class TestGetKerberosPrincipal:

    @mock.patch('subprocess.check_output')
    def test_existing_principal(self, mock_check_output):
        assert get_kerberos_principal_with_keytab(
            'ggroup',
            '/some/keytab',
            'create/admin',
        )

        mock_check_output.assert_called_once_with(
            ['/usr/bin/kadmin', '-K', '/some/keytab', '-p', 'create/admin', 'get', 'ggroup'],
            timeout=10,
            stderr=subprocess.STDOUT,
        )

    @mock.patch('subprocess.check_output')
    def test_nonexistent_principal(self, mock_check_output):
        mock_check_output.side_effect = subprocess.CalledProcessError(
            1,
            ['/usr/bin/kadmin', '-K', '/some/keytab', '-p', 'create/admin', 'get', 'ggroup'],
            b'kadmin: get ggroup: Principal does not exist',
        )

        assert not get_kerberos_principal_with_keytab(
            'ggroup',
            '/some/keytab',
            'create/admin',
        )

    @mock.patch('subprocess.check_output')
    def test_error(self, mock_check_output):
        mock_check_output.side_effect = subprocess.CalledProcessError(
            1,
            ['/usr/bin/kadmin', '-K', '/some/keytab', '-p', 'create/admin', 'get', 'ggroup'],
            b'kadmin: get ggroup: no such file or directory'
        )

        with pytest.raises(ValueError):
            get_kerberos_principal_with_keytab(
                'ggroup',
                '/some/keytab',
                'create/admin',
            )



import mock
import pytest

from ocflib.vhost.web import get_vhost_db
from ocflib.vhost.web import get_vhosts
from ocflib.vhost.web import has_vhost


VHOSTS_EXAMPLE = """
# [added 2015.05.05 ckuehl]
asucarch archive.asuc.org www.archive.asuc.org,modern.asuc.org,www.modern.asuc.org -

# [added 2015.04.16 ckuehl]
staff! contrib - /contrib [nossl]
ocfwiki docs.ocf.berkeley.edu - - [hsts]
"""  # noqa

VHOSTS_EXAMPLE_PARSED = {
    'archive.asuc.org': {
        'aliases': [
            'www.archive.asuc.org',
            'modern.asuc.org',
            'www.modern.asuc.org',
        ],
        'docroot': '/',
        'flags': [],
        'redirect': None,
        'username': 'asucarch',
    },
    'contrib.berkeley.edu': {
        'aliases': [],
        'docroot': '/contrib',
        'flags': ['nossl'],
        'redirect': '/ https://www.ocf.berkeley.edu/~staff/',
        'username': 'staff',
    },
    'docs.ocf.berkeley.edu': {
        'aliases': [],
        'docroot': '/',
        'flags': ['hsts'],
        'redirect': None,
        'username': 'ocfwiki',
    },
}


@pytest.yield_fixture
def mock_get_vhosts_db():
    with mock.patch(
        'ocflib.vhost.web.get_vhost_db',
        return_value=VHOSTS_EXAMPLE.splitlines()
    ):
        yield


class TestVirtualHosts:

    def test_reads_file_if_exists(self):
        with mock.patch('builtins.open', mock.mock_open()) as mock_open:
            lines = ['hello', 'world']
            mock_open.return_value.__iter__.return_value = lines
            assert get_vhost_db() == lines

    @mock.patch('builtins.open')
    @mock.patch('requests.get')
    def test_reads_web_if_no_file(self, get, mock_open):
        def raise_error(__):
            raise IOError()

        mock_open.side_effect = raise_error
        get.return_value.text = 'hello\nworld'

        assert get_vhost_db() == ['hello', 'world']

    def test_proper_parse(self, mock_get_vhosts_db):
        assert get_vhosts() == VHOSTS_EXAMPLE_PARSED

    @pytest.mark.parametrize('user,should_have_vhost', [
        ('staff', True),
        ('ocfwiki', True),
        ('ckuehl', False),
        ('', False),
    ])
    def test_has_vhost(self, user, should_have_vhost, mock_get_vhosts_db):
        assert has_vhost(user) == should_have_vhost
